/*
  GameEngine5.cpp - Arduino Game Engine for a ST7565 based 12864 LCD.
  Created by Thomas Baum, th.baum@online.de, 27.10.2015
*/

#include <util/delay.h>
#include <SPI.h>
#include "GameEngine5.h"
#include "Joypad.h"

Display lcd5 = Display();
Joypad b = Joypad();

byte spawnCount = 10;

GameEngine5::GameEngine5() {
  b.init();
}

void GameEngine5::init(byte contrast) {
  lcd5.init(contrast);
  lcd5.clearVideoBuffer();
}

void GameEngine5::clearVideoBuffer() {
  lcd5.clearVideoBuffer();
}

void GameEngine5::drawBitmap(byte x, byte y, byte width, byte height, const byte *bitmap) {
  lcd5.drawBitmap(x, y, width, height, bitmap);
}

void GameEngine5::invertRectangle(byte x1, byte y1, byte x2, byte y2) {
  lcd5.invertRectangle(x1, y1, x2, y2);
}


void GameEngine5::drawInfo(const char *c) {
  lcd5.fillRectangle(4, 15, 124, 55);
  lcd5.invertRectangle(6, 17, 122, 53);
  lcd5.drawString(8, 30, c, false);
  lcd5.show();
  delay(200);
  while (digitalRead(A)) {
  }
  delay(200);
}

void GameEngine5::show() {

}

void GameEngine5::start() {
  lcd5.show();
}

void GameEngine5::setObjectTileSet(const byte *bitmap) {
  objectTileSet = bitmap;
}

void GameEngine5::setStructureTileSet(const byte *bitmap) {
  structureTileSet = bitmap;
}

void GameEngine5::setDecorationTileSet(const byte *bitmap) {
  decorationTileSet = bitmap;
}

void GameEngine5::setPlayerTile(const byte *bitmap) {
  playerTile = bitmap;
}

void GameEngine5::setStatusPane(const byte *bitmap) {
  isStatusPaneActive = true;
  statusPaneBitmap = bitmap;
}

void GameEngine5::loadLevel(const char *bitmap) {
  levelTileSet = bitmap;
  player.x = 0;
  player.y = 0;
  player.vx = 0;
  player.vy = 0;
  cameraPos = 0;
  for (int i = 0; i < spawnCount; i++) {
    spawn[i].isDrawn = false;
    spawn[i].position = 0;
  }

  clearObjects();
}

void GameEngine5::removeObject(int i) {
  objectArray[i].bitmap = empty;
  objectArray[i].x = 0;
  objectArray[i].vx = 0;
  objectArray[i].px = 0;
  objectArray[i].y = 0;
  objectArray[i].vy = 0;
  objectArray[i].py = 0;

  /*
    for(int i = 0; i<10; i++){
      if (spawn[i].isDrawn && spawn[i].position == levelPosition) {
        return true;
      }
    }
  */
}

void GameEngine5::clearObjects() {
  for (int i = 0; i < objectCount; i++) {
    objectArray[i].bitmap = empty;
    objectArray[i].position = 255;
  }
}

void GameEngine5::addObjectPattern(float vx, float vy, float jx, float jy, boolean gravity, boolean deadly, byte position) {
  objectPattern[position].x = 0;
  objectPattern[position].y = 0;
  objectPattern[position].vx = vx;
  objectPattern[position].vy = vy;
  objectPattern[position].px = vx;
  objectPattern[position].py = vy;
  objectPattern[position].jx = jx;
  objectPattern[position].jy = jy;
  objectPattern[position].gravity = gravity;
  objectPattern[position].deadly = deadly;
  objectPattern[position].bitmap = objectTileSet + (8 * position);
  objectPattern[position].position = position;
}

void GameEngine5::addObject(int x, int y, byte creaturePosition) {
  for (int i = 0; i < objectCount; i++) {
    if (objectArray[i].bitmap == empty) {
      objectArray[i] = objectPattern[creaturePosition];
      objectArray[i].x = x;
      objectArray[i].y = y;
      break;
    }
  }
}

void GameEngine5::updateObjects() {
  for (int i = 0; i < objectCount; i++) {
    if (objectArray[i].bitmap != empty) {
      objectArray[i].x += objectArray[i].vx;
      if (collideObject(objectArray[i].x, objectArray[i].y, objectArray[i].bitmap)) {
        objectArray[i].x -= objectArray[i].vx;
        objectArray[i].vx *= objectArray[i].jx;
      }
      objectArray[i].y += objectArray[i].vy;
      if (collideObject(objectArray[i].x, objectArray[i].y, objectArray[i].bitmap)) {
        objectArray[i].y -= objectArray[i].vy;
        objectArray[i].vy *= objectArray[i].jy;
      }
      if (objectArray[i].gravity) {
        objectArray[i].vy += objectArray[i].py;
        if (objectArray[i].vy > gravity) {
          objectArray[i].vy = gravity;
        }
      }
      if (objectArray[i].y > LCDHEIGHT || objectArray[i].x - cameraPos <= -10 || objectArray[i].x - cameraPos > LCDWIDTH + 10) {
        removeObject(i);
      }
      else {
        lcd5.drawBitmap(objectArray[i].x - cameraPos, objectArray[i].y, 8, 8, objectArray[i].bitmap);
      }
    }
  }
}

boolean GameEngine5::collideObject(float x, float y, const byte* bMap) {
  int approachI = y / 8;
  int approachJ = x / 8;
  for (int i = approachI - 2; i < approachI + 2; i++) {
    for (int j = approachJ - 2; j < approachJ + 2; j++) {
      char character = getLevelTile((i * levelX) + j);
      if (((65 <= character) && (character <= 90)) || character == '#') {
        block = structureTileSet + (character - 65) * 8;
        if (collideBitmapBitmap(x, y, bMap, j * 8, LCDHEIGHT - (8 * (8 - i)), block)) {
          return true;
        }
      }
    }
  }
  return false;
}

char GameEngine5::getLevelTile(int tilePosition) {
  return (char)pgm_read_byte(levelTileSet + tilePosition);
}

void GameEngine5::drawLevel() {
  for (int i = 0; i < 8; i++) {
    for (int j = 0; j < levelX; j++) {
      if ((j * 8 - cameraPos >= -8 && j * 8 - cameraPos <= LCDWIDTH + 8)) {
        block = NULL;
        char character = getLevelTile((i * levelX) + j);
        if ((48 <= character) && (character <= 57)) {
          block = objectTileSet;
          block += (character - 48) * 8;

          if (!enemyExists((word) (i * levelX) + j)) {
            addObject(j * 8, LCDHEIGHT - (8 * (8 - i)), character - 48);
          }

        } else if ((65 <= character) && (character <= 90)) {
          block = structureTileSet;
          block += (character - 65) * 8;
          lcd5.drawBitmap(j * 8 - cameraPos, LCDHEIGHT - (8 * (8 - i)), 8, 8, block);
        } else if ((97 <= character) && (character <= 122)) {
          block = decorationTileSet;
          block += (character - 97) * 8;
          lcd5.drawBitmap(j * 8 - cameraPos, LCDHEIGHT - (8 * (8 - i)), 8, 8, block);
        }
      }
    }
  }
}

boolean GameEngine5::enemyExists(word levelPosition) {
  for (int i = 0; i < spawnCount; i++) {
    if (spawn[i].isDrawn && spawn[i].position == levelPosition) {
      return true;
    }
  }
  for (int i = 0; i < spawnCount; i++) {
    if (!spawn[i].isDrawn) {
      spawn[i].isDrawn = true;
      spawn[i].position = levelPosition;
      return false;
    }
  }

  return true;
}

void GameEngine5::drawPlayer() {
  if (player.invincible % 2 != 1) {
    lcd5.drawBitmap(player.x - cameraPos, player.y, 8, 8, character);
  }
}

void GameEngine5::updatePlayer() {
  walking = false;
  if (player.invincible > 0) {
    player.invincible--;
  }
  player.vx = player.vx * 0.75;
  if (b.repeat(RIGHT, 1)) {
    player.vx += 0.15;
    walking = true;
  }

  if (b.repeat(LEFT, 1)) {
    player.vx -= 0.15;
    walking = true;
  }

  if (frameCount % 10 >= 0 && frameCount % 6 < 3 && walking) {
    character = playerTile;
    player.h = 8;
  }
  else {
    character = playerTile;
    player.h = 8;
  }

  if (player.vx > 1.5) {
    player.vx = 1.5;
  }

  player.x += player.vx;
  if (playerCollide()) {
    player.x -= player.vx;
    player.vx = 0;
  }

  if (player.vy > 1) {
    canJump = false;
  }
  if (canJump && (b.isLongPressed(A) > 0 && b.isLongPressed(A) < 5)) {
    player.vy -= 1.10;
    if (b.isLongPressed(A) < 2) {
      tone(3, 450, 20);
    }
  }
  player.vy += 0.25;
  if (player.vy > 1.5) {
    player.vy = 1.5;
  }

  if (b.isReleased(A)) {
    canJump = false;
  }

  player.y += player.vy;
  if (playerCollide()) {
    if (player.vy > 0) {
      canJump = true;
      for (int i = 0; i < player.vy; i++) {
        player.y -= 1;
        if (!playerCollide()) {
          player.vy = 0;
          break;
        }
      }
    }
    else {
      player.y -= player.vy;
      player.vy /= 2;
    }
  }

  if (player.x + 4 - cameraPos > (int)(LCDWIDTH / 2)  && cameraPos < (levelX * 8) - (LCDWIDTH)) {
    cameraPos = player.x - LCDWIDTH / 2 + 4;
  }

  if (player.x + 4 - cameraPos < (int)(LCDWIDTH / 2) && cameraPos > 0) {
    cameraPos = player.x - LCDWIDTH / 2 + 4;
  }
}

boolean GameEngine5::playerCollide() {
  collide = false;
  for (int i = 0; i < objectCount; i++) {
    if (collideBitmapBitmap(objectArray[i].x, objectArray[i].y, objectArray[i].bitmap, player.x, player.y, character)) {
      fireCollideAction(objectArray[i].position, objectArray[i].deadly);
      if (objectArray[i].deadly) {
        if (player.vy > 0 && player.x - objectArray[i].x <= 7) {
          removeObject(i);
          bounce = true;
          tone(3, 300, 20);
        }
        else {
          if (player.invincible <= 0) {
            objectArray[i].vx *= -1;
          }
        }
      } else {
        removeObject(i);
        //fireCollideAction(0, objectArray[i].deadly);
      }
    }
  }
  int approachI = player.y / 8;
  int approachJ = player.x / 8;
  for (int i = approachI - 2; i < approachI + 2; i++) {
    for (int j = approachJ - 2; j < approachJ + 2; j++) {
      block = NULL;
      char c = getLevelTile((i * levelX) + j);
      if ((65 <= c) && (c <= 90)) {
        block = structureTileSet;
        block += (c - 65) * 8;
      }
      if (!(block == NULL)) {
        if (collideBitmapBitmap(player.x, player.y, character, j * 8, LCDHEIGHT - (8 * (8 - i)), block)) {
          collide = true;
        }
      }
    }
  }
  if (bounce) {
    player.vy = -4;
    bounce = false;
  }
  return collide;
}

void GameEngine5::registerCollideAction( void (*fp)(byte, boolean) ) {
  fpAction = fp;
}

void GameEngine5::fireCollideAction(byte objectID, boolean isDeadly) {
  if ( 0 != fpAction ) {
    (*fpAction)(objectID, isDeadly);
  }
}

void GameEngine5::startPlayer() {
  player.x = 0;
  player.y = 0;
  player.vx = 0;
  player.vy = 0;
  player.h = 4;
  player.invincible = 40;
  player.flip = false;
  player.big = false;
  player.crouch = false;
}

boolean GameEngine5::isFrameCount(byte frame) {

  if ((frameCount % frame) == 0) {

    return true;
  } else {

    return false;
  }
}

boolean GameEngine5::update() {
  if (((nextFrameMillis - millis()) > timePerFrame) && frameEndMicros) {
    nextFrameMillis = millis() + timePerFrame;
    frameCount++;

    frameEndMicros = 0;
    frameStartMicros = micros();
	b.update();
    return true;

  } else {
    if (!frameEndMicros) {
      lcd5.show();
      lcd5.clearVideoBuffer();

      frameEndMicros = micros();
      frameDurationMicros = frameEndMicros - frameStartMicros;
    }
    return false;
  }
}

boolean GameEngine5::isButtonPressed(int button) {
  return !digitalRead(button);
}

boolean GameEngine5::collidePointRect(int16_t x1, int16_t y1 , int16_t x2 , int16_t y2, int16_t w, int16_t h) {
  if ((x1 >= x2) && (x1 < x2 + w))
    if ((y1 >= y2) && (y1 < y2 + h))
      return true;
  return false;
}

boolean GameEngine5::collideRectRect(int16_t x1, int16_t y1, int16_t w1, int16_t h1 , int16_t x2 , int16_t y2, int16_t w2, int16_t h2) {
  if (collidePointRect(x1, y1, x2, y2, w2, h2))
    return true;
  if (collidePointRect(x1 + w1 - 1, y1, x2, y2, w2, h2))
    return true;
  if (collidePointRect(x1, y1 + h1 - 1, x2, y2, w2, h2))
    return true;
  if (collidePointRect(x1 + w1 - 1, y1 + h1 - 1, x2, y2, w2, h2))
    return true;
  return false;
}
boolean GameEngine5::collideBitmapBitmap(int16_t x1, int16_t y1, const uint8_t* b1, int16_t x2, int16_t y2, const uint8_t* b2) {
  int16_t w1 = 8;
  int16_t h1 = 8;
  int16_t w2 = 8;
  int16_t h2 = 8;

  if (collideRectRect(x1, y1, w1, h1, x2, y2, w2, h2) == false) {
    return false;
  }

  int16_t xmin = (x1 >= x2) ? 0 : x2 - x1;
  int16_t ymin = (y1 >= y2) ? 0 : y2 - y1;
  int16_t xmax = (x1 + w1 >= x2 + w2) ? x2 + w2 - x1 : w1;
  int16_t ymax = (y1 + h1 >= y2 + h2) ? y2 + h2 - y1 : h1;
  for (uint8_t y = ymin; y < ymax; y++) {
    for (uint8_t x = xmin; x < xmax; x++) {
      if (getBitmapPixel(b1, x, y) && getBitmapPixel(b2, x1 + x - x2, y1 + y - y2)) {
        return true;
      }
    }
  }
  return false;
}

boolean GameEngine5::getBitmapPixel(const uint8_t* bitmap, uint8_t x, uint8_t y) {
  return pgm_read_byte(bitmap + x) & B00000001 << y;
}

void GameEngine5::updateStatus(int livesValue, int coinsValue, int timeValue) {
   time = timeValue;
   coins = coinsValue;
   lives = livesValue;
}

void GameEngine5::drawStatus() {
  lcd5.drawBitmap(0, 0, 128, 8, statusPaneBitmap);
  
  char buffer[20];
  sprintf(buffer, "%d", lives);
    lcd5.drawString(16, 1, buffer, false);

    sprintf(buffer, "%d", coins);
    lcd5.drawString(56, 1, buffer, false);

    sprintf(buffer, "%d", time);
    lcd5.drawString(100, 1, buffer, false);
}



