/*
  GameEngine5.h - Arduino Game Engine for a ST7565 based 12864 LCD.
  Created by Thomas Baum, th.baum@online.de, 27.10.2015
*/

#ifndef _GameEngine5_H_
#define _GameEngine5_H_

#if ARDUINO >= 100
#include "Arduino.h"
#else
#include "WProgram.h"
#endif

#include "Display.h"
#include "Joypad.h"

struct Char {
  int invincible;
  float x;
  float vx;
  float y;
  float vy;
  byte h;
  boolean flip;
  boolean big;
  boolean crouch;
};

struct ObjectProperty {
  float x;
  float y;
  float vx;
  float vy;
  float px;
  float py;
  float jx;
  float jy;
  boolean gravity;
  boolean deadly;
  const byte *bitmap;
  byte position;
};

struct ObjectSpawn {
  word position;
  boolean isDrawn;
};

const byte empty[] PROGMEM = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

#define objectCount 8

class GameEngine5 {
  public:

    GameEngine5();
    void init(byte contrast);
    void clearVideoBuffer(void);
    void drawBitmap(byte x, byte y, byte width, byte height, const byte bitmap[]);
    void invertRectangle(byte x1, byte y1, byte x2, byte y2);
    void drawInfo(const char *c);
    void show();
    void start();
    boolean isPointOfAction (byte x1, byte y1, byte x2, byte y2);
    void drawCurser();
    boolean isFrameCount(byte frame);
    boolean update();

    void setObjectTileSet(const byte bitmap[]);
    void setStructureTileSet(const byte bitmap[]);
    void setDecorationTileSet(const byte bitmap[]);
    void setStatusPane(const byte bitmap[]);
    void setPlayerTile(const byte bitmap[]);
    void loadLevel(const char bitmap[]);
    void addObjectPattern(float vx, float vy, float jx, float jy, boolean gravity, boolean deadly, byte position);
    void removeObject(int i);
    void clearObjects();
    void addObject(int x, int y, byte position);
    void updateObjects();
    boolean collideObject(float x, float y, const byte* bMap);
    char getLevelTile(int tilePosition);
    void drawLevel();
    boolean enemyExists(word levelPosition);

    void startPlayer();
    void drawPlayer();
    void updatePlayer();
    boolean playerCollide();

    void registerCollideAction( void (*fp)(byte, boolean));
    void fireCollideAction(byte objectID, boolean isDeadly);

    boolean isButtonPressed(int button);
    void updateStatus(int livesValue, int coinsValue, int timeValue);
    void drawStatus();

    boolean getBitmapPixel(const uint8_t* bitmap, uint8_t x, uint8_t y);
    boolean collidePointRect(int16_t x1, int16_t y1 , int16_t x2 , int16_t y2, int16_t w, int16_t h);
    boolean collideRectRect(int16_t x1, int16_t y1, int16_t w1, int16_t h1 , int16_t x2 , int16_t y2, int16_t w2, int16_t h2);
    boolean collideBitmapBitmap(int16_t x1, int16_t y1, const uint8_t* b1, int16_t x2, int16_t y2, const uint8_t* b2);

    void playOK();

    uint32_t frameCount;
    uint16_t frameDurationMicros;
    uint32_t frameStartMicros, frameEndMicros;
    uint8_t timePerFrame;
    uint32_t nextFrameMillis;
    Joypad b;

  private:
    void (*fpAction)(byte, boolean) = 0;

    boolean isStatusPaneActive;
    boolean isNavigationPaneActive;
    boolean isItemPaneActive;
    byte curserPaneX;
    byte curserPaneY;
    byte curserPaneWidth;
    byte curserPaneHeight;
    byte curserX;
    byte curserY;

    ObjectProperty objectArray[objectCount];
    ObjectProperty objectPattern[objectCount];
    ObjectSpawn spawn[10];
    boolean collide;
    const byte *block;
    byte levelX = 128;
    byte levelY = 8;
    byte gravity = 3;

    int cameraPos = 0;
    const byte* objectTileSet;
    const byte* structureTileSet;
    const byte* decorationTileSet;
    const byte* statusPaneBitmap;
    const byte* playerTile;
    const byte *character;
    const char* levelTileSet;

    Char player;
    boolean canJump = false;
    boolean walking = false;
    boolean bounce = false;

    int frames;
    int time;
    int coins;
    int lives;
};

#endif
